#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/soundcard.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch/s3c44b0x_iodirect.h>

#define __author__ "Martin Siu"
#define __description__ "Cirrus Logic CS43L43 Stereo DAC with Headphone Amp"
#define __license__ "GPL"
#define __driver_name__ "cs43143"

//#define CS43L43_DEBUG
#define CS43L43_ENABLE_VOLUME_CONTROL
#define CS43L43_ENABLE_TONE_CONTROL
#define CS43L43_ENABLE_FREQUENCY_CONTROL
#define CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
#define CS43L43_ENABLE_SPEAKER_CONTROL

#define CS43L43_SILENT_BUF_SIZE     0x400    // 1k

#define CS43L43_11KHZ_DATA_BUF_SIZE       0x1000  // 4k
#define CS43L43_11KHZ_DMA_BUF_SIZE        0x400   // 1k

#define CS43L43_16KHZ_DATA_BUF_SIZE       0x1000  // 4k
#define CS43L43_16KHZ_DMA_BUF_SIZE        0x400   // 1k

#define CS43L43_44KHZ_DATA_BUF_SIZE       0x20000  // 128k
#define CS43L43_44KHZ_DMA_BUF_SIZE        0x8000   // 32k

#define CS43L43_I2C_MODE_WRITE      1
#define CS43L43_I2C_MODE_POLL       2
#define CS43L43_I2C_TIMEOUT         0x80000
#define CS43L43_I2C_POLL_INTERVAL   0x80

#define CS43L43_WHEEL_POLL_INTERVAL  (HZ / 10)

struct cs43l43_t
{
    int open_count;
    
    spinlock_t tx_lock;
    
    // ring buffer
    int count;
    int head;
    int tail;
    int data_buf_size;
    int dma_buf_size;
    int dma_kick_start;
    unsigned char data_buffer[CS43L43_44KHZ_DATA_BUF_SIZE]; // we need the largest buffer of all frequencies
    unsigned char silent_buffer[CS43L43_SILENT_BUF_SIZE];
    
    wait_queue_head_t tx_wait_queue;    
    
    int write_seq_no;
    int irq_seq_no;

#ifdef CS43L43_ENABLE_VOLUME_CONTROL        
    unsigned char i2c_data[2];
    int i2c_data_count;
    int i2c_data_ptr;
    unsigned int i2c_status;
    int i2c_mode;

    int wheel_value;
    int wheel_new_value;
    int wheel_elastic;
    struct timer_list wheel_timer;
    
    int volume;
#endif    

#ifdef CS43L43_ENABLE_TONE_CONTROL
    unsigned char bass;
    unsigned char treble;
#endif

#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
    unsigned long pllcon;
#endif    

#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
    int speaker;
#endif    
};

static struct cs43l43_t cs43l43;

#ifdef CS43L43_ENABLE_VOLUME_CONTROL    
static unsigned char cs43l43_volume_map[16] = 
    { 0xe0, 0xe4, 0xe8, 0xec, 
      0xf0, 0xf4, 0xf8, 0xfc, 
      0x00, 0x01, 0x02, 0x03,
      0x04, 0x05, 0x06, 0x07};
#endif      

#ifdef CS43L43_DEBUG
static void cs43l43_dump_data_status (void)
{
    printk("head ptr: 0x%08X, head: 0x%08X, tail: 0x%08X, count: 0x%08X\n", 
            (unsigned int)(cs43l43.data_buffer + cs43l43.head), cs43l43.head, cs43l43.tail, cs43l43.count);
}            
#endif

#ifdef CS43L43_ENABLE_VOLUME_CONTROL    
static void cs43l43_i2c_irq_handler (int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long iiccon, iicstat;
    unsigned char data;
    int i;
    
    iiccon = IO_S3C44B0X_IICCON;
    iicstat = IO_S3C44B0X_IICSTAT;
    
    switch (cs43l43.i2c_mode)
    {
        case CS43L43_I2C_MODE_POLL:
            cs43l43.i2c_status = iicstat;
            break;
            
        case CS43L43_I2C_MODE_WRITE:
            if ((cs43l43.i2c_data_count--) == 0)
            {
                IO_S3C44B0X_IICSTAT=0xd0;	//stop MasTx condition
		        IO_S3C44B0X_IICCON=0xaf;	//resumes IIC operation.
		        udelay(10);	//wait until stop condtion is in effect.
                break;
            }
           
            data = cs43l43.i2c_data[cs43l43.i2c_data_ptr++];
            IO_S3C44B0X_IICDS = data;
            for(i = 0; i < 10; i++);	    //for setup time until rising edge of IICSCL
            IO_S3C44B0X_IICCON = 0xaf;
            break;

        default:
            printk("invalid i2c mode\n");
            break;                 
                        
    }
}

static int cs43l43_i2c_write (unsigned char addr, unsigned char data)
{
    int timeout;
    
    cs43l43.i2c_mode = CS43L43_I2C_MODE_WRITE;
    
    cs43l43.i2c_data_ptr = 0;
    cs43l43.i2c_data[0] = addr;
    cs43l43.i2c_data[1] = data;
    cs43l43.i2c_data_count = 2;
    
    IO_S3C44B0X_IICDS = 0x20;
    IO_S3C44B0X_IICSTAT = 0xf0; //MasTx,Start
    
    udelay(CS43L43_I2C_POLL_INTERVAL);
    timeout = 0;
    while (cs43l43.i2c_data_count != -1)
    {
        udelay(CS43L43_I2C_POLL_INTERVAL);
        timeout += CS43L43_I2C_POLL_INTERVAL;
        if (timeout > CS43L43_I2C_TIMEOUT)
        {
            printk("Timeout error %s()\n", __FUNCTION__);
            break;
        }
    }        
    
    cs43l43.i2c_mode = CS43L43_I2C_MODE_POLL;
    timeout = 0;
    while (1)
    {
	    IO_S3C44B0X_IICDS = 0x20;
	    cs43l43.i2c_status = 0x100;
	    IO_S3C44B0X_IICSTAT = 0xf0; //MasTx,Start
	    IO_S3C44B0X_IICCON = 0xaf;  //resumes IIC operation.
	    
        while (cs43l43.i2c_status == 0x100)
        {
            udelay(CS43L43_I2C_POLL_INTERVAL);
            timeout += CS43L43_I2C_POLL_INTERVAL;
            if (timeout > CS43L43_I2C_TIMEOUT)
            {
                printk("Timeout error %s()\n", __FUNCTION__);
                break;
            }
        }            
        
	    if (!(cs43l43.i2c_status & 0x1))
	        break; // when ACK is received
                        
        if (timeout > CS43L43_I2C_TIMEOUT)
        {
            printk("Timeout error %s()\n", __FUNCTION__);
            break;
        }
    }
    
    IO_S3C44B0X_IICSTAT = 0xd0;  //stop MasTx condition
    IO_S3C44B0X_IICCON = 0xaf;   //resumes IIC operation.
    udelay(64);
    
    return 0;
}

static int cs43l43_poll_wheel (void)
{
    int i;
    int timeout;
    
    IO_S3C44B0X_ADCCON = 0x1 | (0x1 << 2); // AIN1
	
    udelay(8);
    timeout = 0;
    while(!(IO_S3C44B0X_ADCCON & 0x40))
    {

        udelay(CS43L43_I2C_POLL_INTERVAL);
        timeout += CS43L43_I2C_POLL_INTERVAL;
        if (timeout > CS43L43_I2C_TIMEOUT)
        {
            printk("Timeout error %s()\n", __FUNCTION__);
            break;
        }
    }
    
    for (i = 0; i < IO_S3C44B0X_ADCPSR; i++);
    
	i = IO_S3C44B0X_ADCDAT >> 6;
    
    return 0xf - (i & 0xf);
}

static int cs43l43_get_wheel_value (void)
{
    int new_val;
    
    new_val = cs43l43_poll_wheel();    
    if (new_val == cs43l43.wheel_value)
    {
        cs43l43.wheel_elastic = 0;
        cs43l43.wheel_new_value = -1;
    }
    else
    {
        if (cs43l43.wheel_new_value == -1)
        {
            cs43l43.wheel_new_value = new_val;
        }
        else
        {
            if (cs43l43.wheel_new_value == new_val)
            {
                cs43l43.wheel_elastic++;
            }
            else
            {
                // monotonic increase
                if ((new_val > cs43l43.wheel_new_value) && 
                    (cs43l43.wheel_new_value > cs43l43.wheel_value))
                {
                    cs43l43.wheel_elastic++;
                    cs43l43.wheel_new_value = new_val;
                }             
                // monotonic decrease
                else if ((new_val < cs43l43.wheel_new_value) && 
                         (cs43l43.wheel_new_value < cs43l43.wheel_value))
                {
                    cs43l43.wheel_elastic++;
                    cs43l43.wheel_new_value = new_val;
                }                    
                else
                {
                    // treat as noise otherwise
                    cs43l43.wheel_new_value = new_val;
                    cs43l43.wheel_elastic = 0;
                }                    
            }
        }
        
        if (cs43l43.wheel_elastic >= 2)
            cs43l43.wheel_value = cs43l43.wheel_new_value;
    }
    
    return cs43l43.wheel_value;
}

static void cs43l43_wheel_poll_timer (unsigned long data)
{
    int volume;
    
    volume = cs43l43_get_wheel_value();
    if (cs43l43.volume != volume)
    {
        cs43l43.volume = volume;
        cs43l43_i2c_write(0x4, cs43l43_volume_map[volume]);
        cs43l43_i2c_write(0x5, cs43l43_volume_map[volume]);
    }

    // re-queue for next dispatch
    cs43l43.wheel_timer.function = cs43l43_wheel_poll_timer;
    cs43l43.wheel_timer.data = 0;
    cs43l43.wheel_timer.expires = jiffies + CS43L43_WHEEL_POLL_INTERVAL;
    add_timer(&cs43l43.wheel_timer); 
}
#endif // CS43L43_ENABLE_VOLUME_CONTROL

static void cs43l3_queue_dma_buffer (unsigned char *buffer, int size)
{
    unsigned long addr = (unsigned long) buffer;
    
#ifdef CS43L43_DEBUG
    if (buffer != cs43l43.silent_buffer)
        printk("queue buffer: 0x%08X, count: %d\n", (int)buffer, size);
#endif        
    
    // programming bridge dma0 for I2S
    IO_S3C44B0X_BDISRC0 =   (1 << 30) | // 16 bit halfword
                            (1 << 28) | // increment
                            (addr);
                
    IO_S3C44B0X_BDICNT0 =   (1 << 30) | // for I2S
                            (1 << 26) | // unit transfer mode
                            (3 << 22) | // int mode when terminated count
                            (0 << 21) | // disable auto reload
                            (1 << 20) | // enable dma
                            (size);
}

static int cs43l43_config_frequency (int frequency)
{
    unsigned long flags;
    int prescaler;
    int master_clock;
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT    
    unsigned long pllcon;
#endif    
    
    spin_lock_irqsave(&cs43l43.tx_lock, flags);
        
    switch (frequency)
    {
        case 11025:
            cs43l43.data_buf_size = CS43L43_11KHZ_DATA_BUF_SIZE;
            cs43l43.dma_buf_size = CS43L43_11KHZ_DMA_BUF_SIZE;
            prescaler = 0x7;
            master_clock = 1; 
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
            pllcon = 0x3a031; // 66Mhz
#endif
            break;
        
        case 16000:
            cs43l43.data_buf_size = CS43L43_16KHZ_DATA_BUF_SIZE;
            cs43l43.dma_buf_size = CS43L43_16KHZ_DMA_BUF_SIZE;
            prescaler = 0x7;
            master_clock = 0;
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
            pllcon = 0x47041; // 65.83Mhz
#endif            
            break;
        
        case 44100:
            cs43l43.data_buf_size = CS43L43_44KHZ_DATA_BUF_SIZE;
            cs43l43.dma_buf_size = CS43L43_44KHZ_DMA_BUF_SIZE;
            prescaler = 0x2;
            master_clock = 0; 
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
            pllcon = 0x49041; // 67.5Mhz
#endif            
            break;
        
        default:
            spin_unlock_irqrestore(&cs43l43.tx_lock, flags);
            return -EINVAL;    
    }
    
    cs43l43.count = 0;
    cs43l43.head = 0;
    cs43l43.tail = 0;
    memset(cs43l43.data_buffer, 0, cs43l43.data_buf_size);
    
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
    // only update the PLL if it need to be changed.
    if (cs43l43.pllcon != pllcon)
    {
        cs43l43.pllcon = pllcon;
        IO_S3C44B0X_PLLCON = cs43l43.pllcon;
    }        
#endif    
    IO_S3C44B0X_IISPSR = (prescaler << 4) | // Prescaler A for data interface
                         (prescaler << 0);  // Prescaler B for master clock
                       
    if (master_clock)                            
        IO_S3C44B0X_IISMOD |=  (1 << 2); // 384fs for master clock
    else        
        IO_S3C44B0X_IISMOD &= ~(1 << 2); // 256fs for master clock
    
    spin_unlock_irqrestore(&cs43l43.tx_lock, flags);
    
    return 0;
}

static void cs43l43_i2s_init (void)
{
#ifdef CS43L43_DEBUG    
    printk("%s()\n", __FUNCTION__);
#endif    
    
    // I2S disable
    IO_S3C44B0X_IISCON = 0;
    
    IO_S3C44B0X_IISFCON =   (1 << 11) | // Tx FIFO dma mode
                            (1 << 10) | // Rx FIFO dma mode
                            (1 << 9)  | // Tx FIFO enable
                            (0 << 8);   // Rx FIFO disable

    IO_S3C44B0X_IISCON =    (1 << 5) | // Tx dma enable
                            (0 << 4) | // Rx dma disable
                            (0 << 3) | // Tx not idle
                            (0 << 2) | // Rx not idle
                            (1 << 1) | // Prescaler enable
                            (0 << 0);  // I2S disable
                
    IO_S3C44B0X_IISMOD =    (0 << 8) | // Master mode, clock as output
                            (2 << 6) | // Tx mode
                            (0 << 5) | // Low for left, High for right
                            (0 << 4) | // I2S compatible format
                            //(1 << 4) | // Left-justified format
                            (1 << 3) | // 16 bit data
                            (1 << 0);  // 32fs for serial bit clock
                            
    // default for mp3 44kHz                            
    cs43l43_config_frequency(44100);
        
    IO_S3C44B0X_BDIDES0 =   (1 << 30) | // from ext memory to internal periphreal
                            (3 << 28) | // fix address
                            S3C44B0X_IISFIF;
    
    IO_S3C44B0X_BDCON0 =    (0 << 2); // enable dma request
    
    cs43l3_queue_dma_buffer(cs43l43.silent_buffer, CS43L43_SILENT_BUF_SIZE);
    
    IO_S3C44B0X_IISCON |=   (1 << 0); // I2S enable
}

static void cs43l43_hw_init (void)
{
#ifdef CS43L43_DEBUG    
    printk("%s()\n", __FUNCTION__);
#endif      
    
    // Port C 0-3 as IIS function
    IO_S3C44B0X_PCONC |= (0xFF << 0);    
    // Port C 11, 12 as output
    IO_S3C44B0X_PCONC &= ~(0xF << 22);
    IO_S3C44B0X_PCONC |=  (0x5 << 22);
        
    // Port C 0-3 pull up disable
    IO_S3C44B0X_PUPC |= (0xF << 0);
    // Port C 11, 12 pull up disable
    IO_S3C44B0X_PUPC |= (1 << 11) | (1 << 12);
        
    // Port F 0-1 as output
    IO_S3C44B0X_PCONF &= ~(0xF << 0);
    IO_S3C44B0X_PCONF |=  (0x5 << 0);
    
    // Port F 0-1 pull up disable   
    IO_S3C44B0X_PUPF |= (0x3 << 0);
    
    IO_S3C44B0X_PDATF &= ~(0x3 << 0);   // 0x00, I2S format
    //IO_S3C44B0X_PDATF |= (0x1 << 0);  // 0x01  Left justified, up to 24bit data
    //IO_S3C44B0X_PDATF |= (0x3 << 0);  // 0x11  Right justified, 16bit data
    
    IO_S3C44B0X_PDATC &= ~(1 << 12);    // bring RST low, reset DAC
    udelay(200);
    IO_S3C44B0X_PDATC |= (1 << 12); // bring RST high, DAC in operation
}

#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
static void cs43l43_speaker_on (void)
{    
    IO_S3C44B0X_PDATC |= (1 << 11);     // speaker on
}
        
static void cs43l43_speaker_off (void)
{    
    IO_S3C44B0X_PDATC &= ~(1 << 11); // speaker off
}
#endif

static ssize_t cs43l43_write (struct file * file, const char *buf, size_t count, loff_t *ppos)
{
    int free_count, request, written, remaining;
    int n, m;
    unsigned long flags;

    if (count <= 0)
        return 0;

    written = 0;
    remaining = count;
    
    while (remaining > 0)
    {
        while (1)
        {
            spin_lock_irqsave(&cs43l43.tx_lock, flags);
            free_count = cs43l43.data_buf_size - cs43l43.count;

            if (free_count)
                break;

            spin_unlock_irqrestore(&cs43l43.tx_lock, flags);
            interruptible_sleep_on(&cs43l43.tx_wait_queue);
            
            if (signal_pending(current))
                return -ERESTARTSYS;
        }        

#ifdef CS43L43_DEBUG
        printk("before write %d --> ", cs43l43.write_seq_no);        
        cs43l43_dump_data_status();
#endif    
        
        if (remaining > free_count)
            request = free_count;
        else
            request = remaining;        

        n = cs43l43.data_buf_size - cs43l43.tail;
        if (request > n)
        {
#ifdef CS43L43_DEBUG    
            printk("partial write\n");
#endif        
            
            m = request - n;
            memcpy(cs43l43.data_buffer + cs43l43.tail, buf, n); 
            memcpy(cs43l43.data_buffer, buf + n, m); 
            cs43l43.tail = m;
        }
        else
        {
            memcpy(cs43l43.data_buffer + cs43l43.tail, buf, request); 
            cs43l43.tail += request;
            cs43l43.tail %= cs43l43.data_buf_size;
        }
    
        cs43l43.count += request;

#ifdef CS43L43_DEBUG    
        printk("after write %d  --> ", cs43l43.write_seq_no);        
        cs43l43_dump_data_status();
        cs43l43.write_seq_no++;
#endif        
        spin_unlock_irqrestore(&cs43l43.tx_lock, flags);
        
        written += request;
        remaining -= request;
        buf += request;
    }
    
#ifdef CS43L43_DEBUG    
    printk("write audio data: %d\n", written);
#endif    
             
	return written;
}

static void cs43l43_irq_handler (int irq, void *dev_id, struct pt_regs *regs)
{    
    unsigned char *src;
#ifdef CS43L43_DEBUG        
    unsigned char *dst;
    int count;
#endif    

    src = (unsigned char *) (IO_S3C44B0X_BDCSRC0 & 0x0FFFFFFF);
#ifdef CS43L43_DEBUG    
    dst = (unsigned char *) (IO_S3C44B0X_BDCDES0 & 0x0FFFFFFF);
    count = IO_S3C44B0X_BDCCNT0 & 0x000FFFFF;
#endif    
    
    if ((src - CS43L43_SILENT_BUF_SIZE) == cs43l43.silent_buffer)
    {
        // silent buffer done, no need to update ring buffer pointer
    }
    else
    {
#ifdef CS43L43_DEBUG    
        printk("current dma: src: 0x%08X, des: 0x%08X, count: 0x%08X\n", 
            (int)src, (int)dst, (int)count);
        
        printk("before -> ");    
        cs43l43_dump_data_status();
#endif    
        cs43l43.count -= cs43l43.dma_buf_size;
        cs43l43.head += cs43l43.dma_buf_size;
        cs43l43.head %= cs43l43.data_buf_size;

#ifdef CS43L43_DEBUG         
        printk("after  -> ");    
        cs43l43_dump_data_status();        
#endif        
        wake_up_interruptible(&cs43l43.tx_wait_queue);
    }
    
    if (!cs43l43.dma_kick_start)
    {
        if (cs43l43.count >= cs43l43.data_buf_size)
            cs43l43.dma_kick_start = 1;

        cs43l3_queue_dma_buffer(cs43l43.silent_buffer, CS43L43_SILENT_BUF_SIZE);    

        return;
    }
    
    if (cs43l43.count >= cs43l43.dma_buf_size)
    {        
        cs43l3_queue_dma_buffer(cs43l43.data_buffer + cs43l43.head, cs43l43.dma_buf_size);
    }   
    else
    {        
#ifdef CS43L43_DEBUG            
        if (cs43l43.open_count)
            printk("audio buffer underrun\n");
#endif            
        cs43l3_queue_dma_buffer(cs43l43.silent_buffer, CS43L43_SILENT_BUF_SIZE);
    }        
}

static int cs43l43_open (struct inode *inode, struct file *filp)
{
    cs43l43.open_count++;
    cs43l43.dma_kick_start = 0;

    return 0;
}

static int cs43l43_release (struct inode *inode, struct file *filp)
{
    unsigned long flags;
    
    cs43l43.open_count--;
    
    spin_lock_irqsave(&cs43l43.tx_lock, flags);
    
    cs43l43.count = 0;
    cs43l43.head = 0;
    cs43l43.tail = 0;
    // totally clean out the buffer
    memset(cs43l43.data_buffer, 0, cs43l43.data_buf_size);
    
    spin_unlock_irqrestore(&cs43l43.tx_lock, flags);
    
    return 0;
}

static int cs43l43_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    unsigned int lack_data;
#ifdef CS43L43_ENABLE_VOLUME_CONTROL        
    unsigned char volume;
#endif    
#ifdef CS43L43_ENABLE_TONE_CONTROL
    unsigned char tone;
#endif
#ifdef CS43L43_ENABLE_FREQUENCY_CONTROL
    int frequency;
#endif
#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
    int speaker;
#endif    
    
    switch (cmd)
    {
        case SNDCTL_DSP_SYNC:      
            lack_data = cs43l43.count >> 1;
            copy_to_user((void *)arg, &lack_data, sizeof(unsigned int));
            return 0;

#ifdef CS43L43_ENABLE_VOLUME_CONTROL            
        case SOUND_MIXER_WRITE_VOLUME:
            copy_from_user(&volume, (void *)arg, sizeof(unsigned  char));
            cs43l43_i2c_write(0x4, volume);
            udelay(10);
            cs43l43_i2c_write(0x5, volume);
            return 0;
#endif                

#ifdef CS43L43_ENABLE_TONE_CONTROL
        case SOUND_MIXER_WRITE_BASS:
            copy_from_user(&cs43l43.bass, (void *)arg, sizeof(unsigned char));
            tone = cs43l43.bass << 4 | cs43l43.treble;
            cs43l43_i2c_write(0x6, tone);
            return 0;
            
        case SOUND_MIXER_READ_BASS:
            copy_to_user((void *)arg, &cs43l43.bass, sizeof(unsigned char));
            return 0;            
            
        case SOUND_MIXER_WRITE_TREBLE:
            copy_from_user(&cs43l43.treble, (void *)arg, sizeof(unsigned char));
            tone = cs43l43.bass << 4 | cs43l43.treble;
            cs43l43_i2c_write(0x6, tone);
            return 0;
            
        case SOUND_MIXER_READ_TREBLE:
            copy_to_user((void *)arg, &cs43l43.treble, sizeof(unsigned char));
            return 0;                        
#endif            

#ifdef CS43L43_ENABLE_FREQUENCY_CONTROL
        case SNDCTL_DSP_SPEED:
            copy_from_user(&frequency, (void *)arg, sizeof(int));
            return cs43l43_config_frequency(frequency);
#endif

#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
        case SOUND_MIXER_WRITE_SPEAKER:
            copy_from_user(&speaker, (void *)arg, sizeof(int));
            if (speaker)
            {
                cs43l43.speaker = 1;
                cs43l43_speaker_on();
            }                
            else
            {
                cs43l43.speaker = 0;
                cs43l43_speaker_off();
            }
            return 0;
            
        case SOUND_MIXER_READ_SPEAKER:
            copy_to_user((void *)arg, &cs43l43.speaker, sizeof(int));
            return 0;
#endif            
      
        default:        
            return -EINVAL;
    }                
}

static struct file_operations cs43l43_fops = 
{
    write:      cs43l43_write,
    ioctl:      cs43l43_ioctl,
    open:       cs43l43_open,
    release:    cs43l43_release,
};

int __init cs43l43_init (void)
{    
    int err;    
    int i;
    
    cs43l43.count = 0;
    cs43l43.head = 0;
    cs43l43.tail = 0;
    cs43l43.open_count = 0;
    cs43l43.write_seq_no = 0;
    cs43l43.irq_seq_no = 0;
#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
    cs43l43.speaker = 1;
#endif    
#ifdef CS43L43_ENABLE_CPU_FREQ_ADJUSTMENT
    cs43l43.pllcon = 0; // cs43l43_config_frequency will modify this value later on
#endif    
    memset(cs43l43.silent_buffer, 0, CS43L43_SILENT_BUF_SIZE);
    
    spin_lock_init(&cs43l43.tx_lock);    
    init_waitqueue_head(&cs43l43.tx_wait_queue);            

    if ((err = register_chrdev(102, __driver_name__, &cs43l43_fops)))
		panic("Unable to register driver %s, err %d\n", __driver_name__, err);

    if ((err = request_irq(S3C44B0X_INTERRUPT_BDMA0, cs43l43_irq_handler, 0, __driver_name__, NULL)))
        panic("Unable to request interrupt for %s, err %d\n", __driver_name__, err); 

#ifdef CS43L43_ENABLE_VOLUME_CONTROL    
    if ((err = request_irq(S3C44B0X_INTERRUPT_IIC, cs43l43_i2c_irq_handler, 0, __driver_name__, NULL)))
        panic("Unable to request interrupt for %s, err %d\n", __driver_name__, err);         
#endif
        
    cs43l43_hw_init();
    cs43l43_i2s_init();
    
#ifdef CS43L43_ENABLE_VOLUME_CONTROL        
    // Port F 0-1 as i2c
    IO_S3C44B0X_PCONF &= ~(0xF << 0);
    IO_S3C44B0X_PCONF |=  (0xA << 0);
    
    IO_S3C44B0X_IICCON = (1<<7)|(1<<6)|(1<<5)|(0x2);
    IO_S3C44B0X_IICADD = 0x10;   
    IO_S3C44B0X_IICSTAT = 0x10;
    
    // cs43l43 cannot be init some of the time
    // so lets give it a few tries
    for (i = 0; i < 3; i++)
    {
        cs43l43_i2c_write(0x1, 0xd3);
        udelay(8);
        cs43l43_i2c_write(0x1, 0xd1);
        udelay(32);
    }
    
    // initialise volume to -1 rather than 0, 
    // so that it can be updated to the actual value by the wheel
    cs43l43.volume = -1;
    cs43l43.wheel_value = 0;
    cs43l43.wheel_new_value = 0;
    cs43l43.wheel_elastic = 0;  
    
    cs43l43.wheel_timer.function = cs43l43_wheel_poll_timer;
    cs43l43.wheel_timer.data = 0;
    cs43l43.wheel_timer.expires = jiffies + CS43L43_WHEEL_POLL_INTERVAL;
    add_timer(&cs43l43.wheel_timer); 
#endif

#ifdef CS43L43_ENABLE_TONE_CONTROL
    cs43l43.bass = 0;
    cs43l43.treble = 0;
    cs43l43_i2c_write(0xa, 0x29);
    cs43l43_i2c_write(0x6, cs43l43.bass << 4 | cs43l43.treble);
#endif

#ifdef CS43L43_ENABLE_SPEAKER_CONTROL
    cs43l43_speaker_on();
#endif    

	printk(KERN_INFO "%s\n", __description__);

	return 0;
}

__initcall(cs43l43_init);
