#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/poll.h> 
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch/s3c44b0x.h>
#include <asm/arch/s3c44b0x_iodirect.h>

#define __author__ "Martin Siu"
#define __description__ "Function keys for Toymaker video player"
#define __license__ "GPL"
#define __driver_name__ "func_key"

//#define FUNC_KEY_DEBUG_KEY
//#define FUNC_KEY_DEBUG_BACKLIGHT_CONTRAST
//#define FUNC_KEY_ADJUST_BACKLIGHT_CONTRAST

#define FUNC_KEY_BUFFER_SIZE        (4)
#define FUNC_KEY_DEBOUNCE_JIFFY     ((unsigned long) (HZ / 30)) // treat debounce monostable as 30ms
#define FUNC_KEY_REPEAT_JIFFY       ((unsigned long) (HZ / 4)) // auto repeat rate 250ms
#define FUNC_KEY_REPEAT_DEBOUNCE    (3)
#define FUNC_KEY_REPEAT_STEP1       (5)
#define FUNC_KEY_REPEAT_STEP2       (10)
#define FUNC_KEY_POLL_INTERVAL      (HZ / 5)

#define FUNC_KEY_NR_KEY             (5)
#define FUNC_KEY_BIT_STOP           (1 << 0)
#define FUNC_KEY_BIT_PLAY           (1 << 1)
#define FUNC_KEY_BIT_FF             (1 << 2)
#define FUNC_KEY_BIT_REW            (1 << 3)
#define FUNC_KEY_BIT_FUNC           (1 << 4)

#define FUNC_KEY_BIT_MASK           (FUNC_KEY_BIT_STOP | \
                                     FUNC_KEY_BIT_PLAY | \
                                     FUNC_KEY_BIT_FF   | \
                                     FUNC_KEY_BIT_REW  | \
                                     FUNC_KEY_BIT_FUNC)

struct func_key_state_t
{
    int pressed;                            // If the key pressed
    unsigned long jif;                      // how long has key been down for (approx) 
    unsigned long last_sample_jif;          // how long has key been down 
    int consecutive;
    int repeat;
};

static spinlock_t func_key_lock;
static struct timer_list func_key_timer;
static wait_queue_head_t func_key_read_wait_queue;
static struct func_key_state_t func_key_state[FUNC_KEY_NR_KEY]; 

static char ring_buffer[FUNC_KEY_BUFFER_SIZE];
static int ring_buffer_head;
static int ring_buffer_tail;
static int ring_buffer_count;

#ifdef FUNC_KEY_ADJUST_BACKLIGHT_CONTRAST
// hack for backlight and contrast adjustment
static unsigned short func_key_contrast;
static unsigned short func_key_backlight;

static void func_key_adjust_lcd (char keycode)
{
    if (IO_S3C44B0X_PDATG & FUNC_KEY_BIT_FUNC)
        return;

    switch (keycode)
    {
        case 0:
            if (func_key_contrast > 0)
            {
#ifdef FUNC_KEY_DEBUG_BACKLIGHT_CONTRAST            
                printk("contrast: %d -> %d\n", (int)func_key_contrast, (int)(func_key_contrast - 1));
#endif                
                func_key_contrast--;
                outw(func_key_contrast, S3C44B0X_TCMPB2);
            }                
            break;
            
        case 1:
            if (func_key_contrast < 10)
            {
#ifdef FUNC_KEY_DEBUG_BACKLIGHT_CONTRAST
                printk("contrast: %d -> %d\n", (int)func_key_contrast, (int)(func_key_contrast + 1));
#endif                
                func_key_contrast++;                
                outw(func_key_contrast, S3C44B0X_TCMPB2);
            }
            break;                
        
        case 2:
            if (func_key_backlight > 100)
            {
#ifdef FUNC_KEY_DEBUG_BACKLIGHT_CONTRAST
                printk("backlight: %d -> %d\n", (int)func_key_backlight, (int)(func_key_backlight - 100));
#endif                
                func_key_backlight -= 100;
                outw(func_key_backlight, S3C44B0X_TCMPB1);
            }                
            break;
            
        case 3:
            if (func_key_backlight < 2400)
            {
#ifdef FUNC_KEY_DEBUG_BACKLIGHT_CONTRAST
                printk("backlight: %d -> %d\n", (int)func_key_backlight, (int)(func_key_backlight + 100));
#endif                
                func_key_backlight += 100;                
                outw(func_key_backlight, S3C44B0X_TCMPB1);
            }
            break;                                
    }

}
#endif

static void func_key_flush_buffer (void)
{
    ring_buffer_head = 0;
    ring_buffer_tail = 0;
    ring_buffer_count = 0;
}

static int func_key_push_key_code (char keycode)
{
    ring_buffer[ring_buffer_tail] = keycode;
    ring_buffer_tail++;
    ring_buffer_tail %= FUNC_KEY_BUFFER_SIZE;
    
    if (ring_buffer_count < FUNC_KEY_BUFFER_SIZE)
    {
        ring_buffer_count++;
    }
    else
    {
        ring_buffer_head++;
        ring_buffer_head %= FUNC_KEY_BUFFER_SIZE;
    }
    
    return 0;
}

static int func_key_pop_key_code (char *keycode)
{
    if (ring_buffer_count > 0)
    {
        *keycode = ring_buffer[ring_buffer_head];
        ring_buffer_head++;
        ring_buffer_head %= FUNC_KEY_BUFFER_SIZE;
        ring_buffer_count--;
        
        return 0;    
    }
    
    return -ENODATA;
}

static int func_key_open (struct inode *inode, struct file *filp)
{
    unsigned long flags;
    
    MOD_INC_USE_COUNT;
    
    spin_lock_irqsave(&func_key_lock, flags);
    func_key_flush_buffer();
    spin_unlock_irqrestore(&func_key_lock, flags);       
    
    return 0;
}

static int func_key_release (struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
    
    return 0;
}

static ssize_t func_key_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    int read, i;
    unsigned long flags;
    
    spin_lock_irqsave(&func_key_lock, flags);
    if (!ring_buffer_count)
    {
        spin_unlock_irqrestore(&func_key_lock, flags);
        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN;
        else
            interruptible_sleep_on(&func_key_read_wait_queue);
    }
    
    spin_lock_irqsave(&func_key_lock, flags);
    if (count < ring_buffer_count)
        read = count;
    else
        read = ring_buffer_count;
        
    for (i = 0; i < read; i++)
        func_key_pop_key_code(buf + i);
    spin_unlock_irqrestore(&func_key_lock, flags);
    
    return read;
}

static void func_key_check_key_press (unsigned long keys)
{
    int keycode, keybit;
    int repeat_jiffy;
         
    for (keycode = 0; keycode < FUNC_KEY_NR_KEY; keycode++)
    {
        keybit = 0x1 << keycode;
        if (!(keys & keybit))
        {
            if (func_key_state[keycode].pressed)
            {
                if (jiffies - func_key_state[keycode].jif > FUNC_KEY_DEBOUNCE_JIFFY)
                {               
                    if (func_key_state[keycode].last_sample_jif == func_key_state[keycode].jif)
                    {
                        // want to detect first key pressed
                        // make it a bit different so that above condition wont be true for next key press;
                        func_key_state[keycode].last_sample_jif++;

                        #define __KEY_PRESSED__
#ifdef FUNC_KEY_DEBUG_KEY
                        printk("key: 0x%X pressed\n", keybit);
#endif                        
                        func_key_state[keycode].repeat++;
                        
#ifdef FUNC_KEY_ADJUST_BACKLIGHT_CONTRAST                        
                        func_key_adjust_lcd(keycode);
#endif                        
                        
                        func_key_push_key_code(keycode);
                        func_key_state[keycode].consecutive++;
                        wake_up_interruptible(&func_key_read_wait_queue);
                    }                            

                    // auto repeat 
                    repeat_jiffy = FUNC_KEY_REPEAT_JIFFY;
                    if (func_key_state[keycode].consecutive > FUNC_KEY_REPEAT_STEP2)
                        repeat_jiffy = FUNC_KEY_REPEAT_JIFFY >> 2;                   
                    else if (func_key_state[keycode].consecutive > FUNC_KEY_REPEAT_STEP1)
                        repeat_jiffy = FUNC_KEY_REPEAT_JIFFY >> 1;                                   
                        
                    if (jiffies - func_key_state[keycode].last_sample_jif > repeat_jiffy)
                    {
                        func_key_state[keycode].last_sample_jif = jiffies;

                        func_key_state[keycode].repeat++;
                        
                        if (func_key_state[keycode].repeat > FUNC_KEY_REPEAT_DEBOUNCE)
                        {
                            #define __KEY_REPEAT__
#ifdef FUNC_KEY_DEBUG_KEY                            
                            printk("key: 0x%X repeat\n", keybit);
#endif                            
                            
                            func_key_push_key_code(keycode);                
                            func_key_state[keycode].consecutive++;    
                            wake_up_interruptible(&func_key_read_wait_queue);
                        }
                    }
                }
                else
                {
                    #define __KEY_DEBOUNCE__
#ifdef FUNC_KEY_DEBUG_KEY                    
                    printk("ignore key: 0x%X debounce\n", keybit);  
#endif                    
                }                        
            }
            else
            {
                func_key_state[keycode].pressed = 1;
                func_key_state[keycode].jif = jiffies;
                func_key_state[keycode].last_sample_jif = jiffies;
            }
        }
    }        
}

static void func_key_check_key_release (unsigned long keys)
{
    int keycode, keybit;

    for (keycode = 0; keycode < FUNC_KEY_NR_KEY; keycode++)
    {
        keybit = 0x1 << keycode;

        if (keys & keybit)
        {
            if (func_key_state[keycode].pressed)
            {
                func_key_state[keycode].pressed = 0;

                // check if we have sampled a key pressed at least once?
                if (func_key_state[keycode].last_sample_jif != func_key_state[keycode].jif)
                {
                    func_key_state[keycode].repeat = 0;
                
                    #define __KEY_RELEASED__
#ifdef FUNC_KEY_DEBUG_KEY
                    printk("key: 0x%X released\n", keybit);
#endif                    
                    func_key_state[keycode].consecutive = 0;
                }         
            }
        }
    }
}

static void func_key_poll_timer (unsigned long data)
{
    unsigned long pdatg;
        
    pdatg = IO_S3C44B0X_PDATG & FUNC_KEY_BIT_MASK;
    
    // there must be at least one key down
    if (pdatg != FUNC_KEY_BIT_MASK)
        func_key_check_key_press(pdatg);

    func_key_check_key_release(pdatg);

    add_timer(&func_key_timer);
}

static unsigned int func_key_poll (struct file *filp, struct poll_table_struct *wait)
{
    unsigned int mask = 0;
    unsigned long flags;
    
    poll_wait(filp, &func_key_read_wait_queue, wait);
    
    spin_lock_irqsave(&func_key_lock, flags);
    if (ring_buffer_count > 0)
        mask = POLLIN | POLLRDNORM;
    spin_unlock_irqrestore(&func_key_lock, flags);        
    
    return mask;
}

static struct file_operations func_key_fops = 
{
    open:       func_key_open,
    release:    func_key_release,
    read:       func_key_read,
    poll:       func_key_poll,
};

static void func_key_hardware_init (void)
{    
    // set as input port
    IO_S3C44B0X_PCONG &=   ~((3 << 0) |  
                             (3 << 2) |  
                             (3 << 4) |  
                             (3 << 6) |
                             (3 << 8));  
                                            
    IO_S3C44B0X_PUPG = 0x00; // pull up enable
}

int __init func_key_init (void)
{    
    int err;    
    
    if ((err = register_chrdev(103, __driver_name__, &func_key_fops)))
		panic("Unable to register driver %s, err %d\n", __driver_name__, err);    
    
    init_waitqueue_head(&func_key_read_wait_queue); 
    func_key_timer.function = func_key_poll_timer;
    func_key_timer.data = 0;
    func_key_timer.expires = jiffies + FUNC_KEY_POLL_INTERVAL;
    add_timer(&func_key_timer); 
    
    func_key_flush_buffer();
    spin_lock_init(&func_key_lock);
    
#ifdef FUNC_KEY_ADJUST_BACKLIGHT_CONTRAST    
    func_key_backlight = IO_S3C44B0X_TCMPB1;
    func_key_contrast = IO_S3C44B0X_TCMPB2;
#endif    
    
    func_key_hardware_init();

	printk(KERN_INFO "%s\n", __description__);

	return 0;
}

__initcall(func_key_init);
