/********************************************
 *
 * File: drivers/char/gpm425a0.c
 * 
 * Description: LCD driver for GiantPlus GPM425A0 4K color STN module
 *
 * Author: Martin Siu, EmSoft Ltd
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/s3c44b0x_lcd.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/s3c44b0x.h>
#include <asm/arch/s3c44b0x_iodirect.h>
#include <asm/arch/memory.h>

#define __author__      "Martin Siu"
#define __description__ "Giantplus GPM425A0 240x160 4096 color STN"
#define __license__     "GPL"
#define __driver_name__ "cstn"

#define GPM425A0_MAJOR      101

//#define GPM425A0_DEBUG

#ifdef GPM425A0_DEBUG
#define dprintk(a...)		printk(##a)
#else
#define dprintk(a...)		while(0){}
#endif

//#define GPM425A0_DRAW_TEST_PATTERN
#define GPM425A0_DRAW_LOGO

#ifdef GPM425A0_DRAW_LOGO
#include "noah_logo.h"
#endif

#define LCD_XSIZE 720
#define LCD_YSIZE 160

#define SCR_XSIZE LCD_XSIZE
#define SCR_YSIZE LCD_YSIZE

#define M5D(n) ((n) & 0x1fffff)

#define HOZVAL			        (LCD_XSIZE / 4 - 1)
#define HOZVAL_COLOR	    	(LCD_XSIZE / 8 - 1)
#define LINEVAL			        (LCD_YSIZE - 1)
#define MVAL			        (13)

#define S3C44B0X_FB_ADDRESS     VIDEO_MEM   // calculated and defined in asm/arch/memory.h

struct gpm425a0_t
{
    unsigned long fb_addr;
};

static struct gpm425a0_t gpm425a0;

static void __init gpm425a0_init_pwm (struct gpm425a0_t *device)
{
    dprintk("%s()\n", __FUNCTION__);
    
    // PE3 for TOUT0
    IO_S3C44B0X_PCONE &= ~(3 << 6);
    IO_S3C44B0X_PCONE |=  (2 << 6);
    
    // PE4 for TOUT1
    IO_S3C44B0X_PCONE &= ~(3 << 8);
    IO_S3C44B0X_PCONE |=  (2 << 8);
    
    // disable timer 0, 1
    IO_S3C44B0X_TCON &= ~(0xF << 0);
    IO_S3C44B0X_TCON &= ~(0xF << 8);
    
    // contrast PWM, each cycle is 2000Hz
    IO_S3C44B0X_TCNTB0 = 2000;
    IO_S3C44B0X_TCMPB0 = 100;
    
    // backlight PWM, each cycle is 2000Hz
    IO_S3C44B0X_TCNTB1 = 2000;
    IO_S3C44B0X_TCMPB1 = 100;
        
    // manual update timer 0, 1
    IO_S3C44B0X_TCON |= (1 << 1);
    IO_S3C44B0X_TCON |= (1 << 9);
    
    // turn off manual update on timer 0, 1
    IO_S3C44B0X_TCON &= ~(1 << 1);
    IO_S3C44B0X_TCON &= ~(1 << 9);
    
    // start timer 0,1 interval mode, auto reload
    IO_S3C44B0X_TCON |= (0x9 << 0);
    IO_S3C44B0X_TCON |= (0x9 << 8);
}

#ifdef GPM425A0_DRAW_TEST_PATTERN
static void __init gpm425a0_draw_test_pattern (struct gpm425a0_t *device) 
{
    unsigned char *addr;
	int i;

    dprintk("%s()\n", __FUNCTION__);

    // disable lcd controller
	IO_S3C44B0X_LCDCON1 &= ~0x1;
    
	addr = (unsigned char *)device->fb_addr;
	for (i = 0; i <120*40 ; i++) {
		*addr = 0xF0;
		addr++;
		*addr = 0x0F;
		addr++;
		*addr = 0x00;
		addr++;
	}
	for (i = 0; i <120*40 ; i++) {
		*addr = 0x0F;
		addr++;
		*addr = 0x00;
		addr++;
		*addr = 0xF0;
		addr++;
	}
	for (i = 0; i <120*40 ; i++) {
		*addr = 0x00;
		addr++;
		*addr = 0xF0;
		addr++;
		*addr = 0x0F;
		addr++;
	}
	for (i = 0; i <120*20 ; i++) {
		*addr = 0xFF;
		addr++;
		*addr = 0xFF;
		addr++;
		*addr = 0xFF;
		addr++;
	}
	for (i = 0; i <120*20 ; i++) {
		*addr = 0x00;
		addr++;
		*addr = 0x00;
		addr++;
		*addr = 0x00;
		addr++;
	}
	
    // enable lcd controller
	IO_S3C44B0X_LCDCON1 |= 0x1;
}
#endif

static void __init gpm425a0_hw_init (struct gpm425a0_t *device)
{
    int clkval, wlh, wdly, mmode, dismode, invclk, invline, invframe, invvd, envid;
    int modesel;
    unsigned long lcdcon1, lcdcon2, lcdaddr1, lcdaddr2, lcdaddr3;

    dprintk("%s()\n", __FUNCTION__);

    // enable clock into LCD controller
    IO_S3C44B0X_CLKCON |= (1 << 6);

    // port D for video
    IO_S3C44B0X_PCOND = 0xAAAA;

    // port D pull up disable
    IO_S3C44B0X_PUPD = 0xFF;

    // GPC[4:7] => VD[7:4]
    IO_S3C44B0X_PCONC |= (0xFF << 8); 

    // GPC8 output (DISP-ON)
    IO_S3C44B0X_PCONC &= ~(3 << 16);
    IO_S3C44B0X_PCONC |=  (1 << 16);
    IO_S3C44B0X_PDATC |=  (1 << 8);

    clkval      = 13;
    wlh         = 3; // 16 clocks
    wdly        = 3; // 16 clocks
    mmode       = 0; 
    dismode     = 2; // 8-bit single scan
    invclk      = 0;
    invline     = 0;
    invframe    = 0;
    invvd       = 0;
    envid       = 0;
    modesel     = 2; // 16-level gray mode
    
    lcdcon1 =   (clkval     << 12) |
                (wlh        << 10) |
                (wdly       <<  8) |
                (mmode      <<  7) |
                (dismode    <<  5) |
                (invclk     <<  4) |
                (invline    <<  3) |
                (invframe   <<  2) |
                (invvd      <<  1) |
                (envid      <<  0);
    
	lcdcon2 = (LINEVAL) | (HOZVAL_COLOR << 10)| (10 << 21);
	lcdaddr1 = (modesel << 27) | ((device->fb_addr >> 22)  << 21) | M5D(device->fb_addr >> 1);
	lcdaddr2 = M5D(((device->fb_addr + (SCR_XSIZE / 2 * LCD_YSIZE)) >> 1)) | (MVAL << 21) | (1 << 29);
	lcdaddr3 = (LCD_XSIZE / 4) | (((SCR_XSIZE - LCD_XSIZE) / 2) << 9);
    
    outl(lcdcon1, S3C44B0X_LCDCON1);
    outl(lcdcon2, S3C44B0X_LCDCON2);
    outl(lcdaddr1, S3C44B0X_LCDSADDR1);
    outl(lcdaddr2, S3C44B0X_LCDSADDR2);
    outl(lcdaddr3, S3C44B0X_LCDSADDR3);

	//The following value has to be changed for better display.
	IO_S3C44B0X_REDLUT = 0xFDB96420;
	IO_S3C44B0X_GREENLUT = 0xFDB96420;
    IO_S3C44B0X_BLUELUT = 0xFB40;
    
	IO_S3C44B0X_DITHMODE    = 0x0;
	IO_S3C44B0X_DP1_2       = 0xA5A5;
	IO_S3C44B0X_DP4_7       = 0xBA5DA65;
	IO_S3C44B0X_DP3_5       = 0xA5A5f;
	IO_S3C44B0X_DP2_3       = 0xD6B;
	IO_S3C44B0X_DP5_7       = 0xEB7B5ED;
	IO_S3C44B0X_DP3_4       = 0x7DBE;
	IO_S3C44B0X_DP4_5       = 0x7EBDF;
	IO_S3C44B0X_DP6_7       = 0x7FDFBFE;

    // enable LCD controller
	IO_S3C44B0X_LCDCON1 |= 1;
}

static void gpm425a0_hw_cleanup (struct gpm425a0_t *device)
{
    dprintk("%s()\n", __FUNCTION__);
    
    // disable LCD controller
    IO_S3C44B0X_LCDCON1 &= ~(1 << 0);
    
    // DISP-OFF
    IO_S3C44B0X_PDATC &= ~(1 << 8);
    
    // disable clock into LCD controller
    IO_S3C44B0X_CLKCON &= ~(1 << 6);
}

static int gpm425a0_open (struct inode *inode, struct file *filp)
{
    filp->private_data = &gpm425a0;
    
    dprintk("%s()\n", __FUNCTION__);
    
    MOD_INC_USE_COUNT;
    
    return 0;
}

static int gpm425a0_release (struct inode *inode, struct file *filp)
{
    dprintk("%s()\n", __FUNCTION__);

    MOD_DEC_USE_COUNT;
    
    return 0;
}

static int gpm425a0_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    struct gpm425a0_t *device = (struct gpm425a0_t *)filp->private_data;

    dprintk("%s()\n", __FUNCTION__);
    
    switch (cmd) {
        case S3C44B0X_LCD_IOC_GET_FRAMEBUFFER:      
            if (copy_to_user((void *)arg, &device->fb_addr, sizeof(unsigned long)))
                return -EFAULT;                
            return 0;
                
        default:        
            return -EINVAL;
    }                
}

static ssize_t gpm425a0_write (struct file *filp, const char *buf, size_t count, loff_t *ppos)
{
    struct gpm425a0_t *device = (struct gpm425a0_t *)filp->private_data;
    int size;
    
    dprintk("%s()\n", __FUNCTION__);
    
    if ((count + *ppos) > VIDEO_SIZE)
        size = VIDEO_SIZE - *ppos;
    else
        size = count;
    
    if (size == 0)
        return -ENOSPC;
                
	copy_from_user((unsigned char *)device->fb_addr + *ppos, buf, size);
    *ppos += size;
    
    return size;
}

static struct file_operations gpm425a0_fops =
{
    open:       gpm425a0_open,
    release:    gpm425a0_release,
    write:      gpm425a0_write,
    ioctl:      gpm425a0_ioctl,
};

static int __init s3c44b0x_lcd_init (void)
{
    struct gpm425a0_t *device = &gpm425a0;
    
    device->fb_addr = S3C44B0X_FB_ADDRESS;
	gpm425a0_hw_init(device);
	gpm425a0_init_pwm(device);

	if (register_chrdev(GPM425A0_MAJOR, __driver_name__, &gpm425a0_fops))
		panic("Unable to register %s\n", __driver_name__);
	    
#ifdef GPM425A0_DRAW_TEST_PATTERN
    gpm425a0_draw_test_pattern(device);
#else
    // fill black
    memset((char *)device->fb_addr, 0x00, VIDEO_SIZE);
#endif

#ifdef GPM425A0_DRAW_LOGO
    memcpy((char *)device->fb_addr, noah_logo, VIDEO_SIZE);
#endif

	printk(KERN_INFO "%s frame buffer address: 0x%08lx\n", __description__, device->fb_addr);

	return 0;
}

static void __init s3c44b0x_lcd_exit (void)
{
    struct gpm425a0_t *device = &gpm425a0;
    
    unregister_chrdev(GPM425A0_MAJOR, __driver_name__); 
    gpm425a0_hw_cleanup(device);
}

MODULE_AUTHOR(__author__);
MODULE_DESCRIPTION(__description__);
MODULE_LICENSE(__license__);

module_init(s3c44b0x_lcd_init);
module_exit(s3c44b0x_lcd_exit);
