#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/interrupt.h> 
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/irqs.h>
#include <asm/arch/s3c44b0x.h>
#include <asm/arch/s3c44b0x_iodirect.h>

#define __author__      "Martin Siu"
#define __description__ "8x8 Matrix keyboard for Noah Dictionary"
#define __license__     "GPL"
#define __driver_name__ "key"

//#define NOAH_KEY_DEBUG

#ifdef NOAH_KEY_DEBUG
#define dprintk(a...)   printk(a)
#else
#define dprintk(a...)   while(0){}
#endif

#define NOAH_KEY_MAJOR              (103)
#define NOAH_KEY_NR_COL             (8)
#define NOAH_KEY_NR_ROW             (8)
#define NOAH_KEY_NR_KEY             (NOAH_KEY_NR_COL * NOAH_KEY_NR_ROW)
#define NOAH_KEY_COL_SCAN_DELAY     (4)     // micro-sec
#define NOAH_KEY_POLL_INTERVAL      (2)     // jiffies
#define NOAH_KEY_DEBOUNCE_INTERVAL  (4)     // jiffies
#define NOAH_KEY_BUF_SIZE           (4)     // will buffer 4 keys at most

struct col_output_t
{
    void (*hi) (void);
    void (*lo) (void);
};

struct row_input_t
{ 
    int (*val) (void);
};

struct state_t
{
    int pressed;
    int debounce;
    unsigned long jif;
};

struct ring_buf_t
{
    int count;
    int head;
    int tail;
    unsigned char data[NOAH_KEY_BUF_SIZE];
};

struct noah_key_t 
{
    int col;
    int row;
    struct state_t state[NOAH_KEY_NR_COL][NOAH_KEY_NR_ROW];
    void (*hw_config) (void);
    struct col_output_t col_output[NOAH_KEY_NR_COL];
    struct row_input_t row_input[NOAH_KEY_NR_ROW];
    
    struct timer_list poll_timer;
    struct ring_buf_t buf;
    spinlock_t lock;
    wait_queue_head_t wait_queue;
};

static struct noah_key_t noah_key;

static void col0_hi (void) { IO_S3C44B0X_PDATB |= (1 << 4);  }
static void col1_hi (void) { IO_S3C44B0X_PDATB |= (1 << 5);  }
static void col2_hi (void) { IO_S3C44B0X_PDATB |= (1 << 7);  }
static void col3_hi (void) { IO_S3C44B0X_PDATB |= (1 << 9);  }
static void col4_hi (void) { IO_S3C44B0X_PDATB |= (1 << 10); }
static void col5_hi (void) { IO_S3C44B0X_PDATF |= (1 << 2);  }
static void col6_hi (void) { IO_S3C44B0X_PDATF |= (1 << 4);  }
static void col7_hi (void) { IO_S3C44B0X_PDATF |= (1 << 3);  }

static void col0_lo (void) { IO_S3C44B0X_PDATB &= ~(1 << 4);  }
static void col1_lo (void) { IO_S3C44B0X_PDATB &= ~(1 << 5);  }
static void col2_lo (void) { IO_S3C44B0X_PDATB &= ~(1 << 7);  }
static void col3_lo (void) { IO_S3C44B0X_PDATB &= ~(1 << 9);  }
static void col4_lo (void) { IO_S3C44B0X_PDATB &= ~(1 << 10); }
static void col5_lo (void) { IO_S3C44B0X_PDATF &= ~(1 << 2);  }
static void col6_lo (void) { IO_S3C44B0X_PDATF &= ~(1 << 4);  }
static void col7_lo (void) { IO_S3C44B0X_PDATF &= ~(1 << 3);  }

static int row0_val (void) { return IO_S3C44B0X_PDATC & (1 << 1)  ? 1 : 0; }
static int row1_val (void) { return IO_S3C44B0X_PDATC & (1 << 2)  ? 1 : 0; }
static int row2_val (void) { return IO_S3C44B0X_PDATC & (1 << 3)  ? 1 : 0; }
static int row3_val (void) { return IO_S3C44B0X_PDATC & (1 << 10) ? 1 : 0; }
static int row4_val (void) { return IO_S3C44B0X_PDATC & (1 << 11) ? 1 : 0; }
static int row5_val (void) { return IO_S3C44B0X_PDATC & (1 << 12) ? 1 : 0; }
static int row6_val (void) { return IO_S3C44B0X_PDATC & (1 << 13) ? 1 : 0; }
static int row7_val (void) { return IO_S3C44B0X_PDATC & (1 << 14) ? 1 : 0; }

static void noah_key_hw_config (void)
{
    dprintk("%s()\n", __FUNCTION__);
    
    IO_S3C44B0X_PCONB &= ~(1 << 4);  // key col1 output
    IO_S3C44B0X_PCONB &= ~(1 << 5);  // key col2 output
    IO_S3C44B0X_PCONB &= ~(1 << 7);  // key col3 output
    IO_S3C44B0X_PCONB &= ~(1 << 9);  // key col4 output
    IO_S3C44B0X_PCONB &= ~(1 << 10); // key col5 output
    IO_S3C44B0X_PCONF &= ~(3 << 4); 
    IO_S3C44B0X_PCONF |=  (1 << 4);  // key col6 output
    IO_S3C44B0X_PCONF &= ~(3 << 8); 
    IO_S3C44B0X_PCONF |=  (1 << 8);  // key col7 output
    IO_S3C44B0X_PCONF &= ~(3 << 6); 
    IO_S3C44B0X_PCONF |=  (1 << 6);  // key col8 output
    
    IO_S3C44B0X_PCONC &= ~(3 << 2);  // key row9 input
    IO_S3C44B0X_PCONC &= ~(3 << 4);  // key row10 input
    IO_S3C44B0X_PCONC &= ~(3 << 6);  // key row11 input
    IO_S3C44B0X_PCONC &= ~(3 << 20); // key row12 input
    IO_S3C44B0X_PCONC &= ~(3 << 22); // key row13 input
    IO_S3C44B0X_PCONC &= ~(3 << 24); // key row14 input
    IO_S3C44B0X_PCONC &= ~(3 << 26); // key row15 input
    IO_S3C44B0X_PCONC &= ~(3 << 28); // key row16 input

    IO_S3C44B0X_PUPC |= (1 << 1);
    IO_S3C44B0X_PUPC |= (1 << 2);
    IO_S3C44B0X_PUPC |= (1 << 3);
    IO_S3C44B0X_PUPC |= (1 << 10);
    IO_S3C44B0X_PUPC |= (1 << 11);
    IO_S3C44B0X_PUPC |= (1 << 12);
    IO_S3C44B0X_PUPC |= (1 << 13);
    IO_S3C44B0X_PUPC |= (1 << 14);
}

static void noah_key_flush_buffer (struct ring_buf_t *buf)
{
    buf->head = 0;
    buf->tail = 0;
    buf->count = 0;
}

static int noah_key_push_key_code (struct ring_buf_t *buf, char keycode)
{
    dprintk("%s()\n", __FUNCTION__);
    dprintk("keycode: %d\n", keycode);
    
    buf->data[buf->tail] = keycode;
    buf->tail++;
    buf->tail %= NOAH_KEY_BUF_SIZE;
    
    if (buf->count < NOAH_KEY_BUF_SIZE) {
        buf->count++;
    } else {
        buf->head++;
        buf->head %= NOAH_KEY_BUF_SIZE;
    }
    
    return 0;
}

static int noah_key_pop_key_code (struct ring_buf_t *buf, char *keycode)
{
    dprintk("%s()\n", __FUNCTION__);

    if (buf->count > 0) {
        *keycode = buf->data[buf->head];
        buf->head++;
        buf->head %= NOAH_KEY_BUF_SIZE;
        buf->count--;
        dprintk("keycode: %d\n", *keycode);
        
        return 0;
    }
    
    dprintk("buffer empty\n");
    return -ENODATA;
}

static void noah_key_scan (struct noah_key_t *key)
{
    int delta_jif;
    unsigned char keycode;
    struct state_t *state;
        
    for (key->col = 0; key->col < NOAH_KEY_NR_COL; key->col++) {
        key->col_output[key->col].hi();
        udelay(NOAH_KEY_COL_SCAN_DELAY);
        
        for (key->row = 0; key->row < NOAH_KEY_NR_ROW; key->row++) {            
            state = &key->state[key->col][key->row];
            
            // sampling...            
            if (key->row_input[key->row].val()) {
                if (!state->debounce) {
                    state->jif = jiffies;
                    state->debounce = 1;
                    dprintk("(%d, %d) : start debounce\n", key->col, key->row);
                } else {
                    delta_jif = jiffies - state->jif;
                    if (delta_jif > NOAH_KEY_DEBOUNCE_INTERVAL) {
                        if (!state->pressed) {
                            dprintk("(%d, %d) : down\n", key->col, key->row);
                            state->pressed = 1;
                            keycode = (key->col * NOAH_KEY_NR_ROW) + key->row;
                            noah_key_push_key_code(&key->buf, keycode);
                            wake_up_interruptible(&key->wait_queue);
                        }
                    } else {
                        dprintk("(%d, %d) : debounce, delta jif: %d\n", key->col, key->row, delta_jif);
                    }
                }
            } else {
                if (state->debounce) {
                    if (state->pressed) {
                        dprintk("(%d, %d) : up\n", key->col, key->row);
                        state->pressed = 0;
                    }
                    state->debounce = 0;
                }
            }
        }
        
        key->col_output[key->col].lo();
    }
}

static void noah_key_timer_dispatch (unsigned long data)
{
    struct noah_key_t *key = (struct noah_key_t *) data;
    
    noah_key_scan(key);
    key->poll_timer.expires = jiffies + NOAH_KEY_POLL_INTERVAL;
    add_timer(&key->poll_timer);
}

static int noah_key_open (struct inode *inode, struct file *filp)
{
    struct noah_key_t *key = &noah_key;
    unsigned long flags;
    
    MOD_INC_USE_COUNT;
    
    filp->private_data = key;
    
    spin_lock_irqsave(&key->lock, flags);
    noah_key_flush_buffer(&key->buf);
    spin_unlock_irqrestore(&key->lock, flags);       
    
    return 0;
}

static int noah_key_release (struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
    
    return 0;
}

static ssize_t noah_key_read (struct file *filp, char *buf, size_t count, loff_t *ppos)
{
    struct noah_key_t *key = (struct noah_key_t *) filp->private_data;
    int read, i;
    unsigned long flags;
    
    spin_lock_irqsave(&key->lock, flags);
    if (!key->buf.count) {
        spin_unlock_irqrestore(&key->lock, flags);
        if (filp->f_flags & O_NONBLOCK)
            return -EAGAIN;
        else
            interruptible_sleep_on(&key->wait_queue);
    }
    
    spin_lock_irqsave(&key->lock, flags);
    if (count < key->buf.count)
        read = count;
    else
        read = key->buf.count;
        
    for (i = 0; i < read; i++)
        noah_key_pop_key_code(&key->buf, buf + i);
    spin_unlock_irqrestore(&key->lock, flags);
    
    return read;

    return 0;
}

static unsigned int noah_key_poll (struct file *filp, struct poll_table_struct *wait)
{
    struct noah_key_t *key = (struct noah_key_t *) filp->private_data;
    unsigned int mask = 0;
    unsigned long flags;
    
    poll_wait(filp, &key->wait_queue, wait);
    
    spin_lock_irqsave(&key->lock, flags);
    if (key->buf.count > 0)
        mask = POLLIN | POLLRDNORM;
    spin_unlock_irqrestore(&key->lock, flags);        
    
    return mask;
}

static void __init noah_key_ctx_init (struct noah_key_t *key)
{    
    dprintk("%s()\n", __FUNCTION__);
    
    key->col_output[0].hi = col0_hi;
    key->col_output[1].hi = col1_hi;
    key->col_output[2].hi = col2_hi;
    key->col_output[3].hi = col3_hi;
    key->col_output[4].hi = col4_hi;
    key->col_output[5].hi = col5_hi;
    key->col_output[6].hi = col6_hi;
    key->col_output[7].hi = col7_hi;
    
    key->col_output[0].lo = col0_lo;
    key->col_output[1].lo = col1_lo;
    key->col_output[2].lo = col2_lo;
    key->col_output[3].lo = col3_lo;
    key->col_output[4].lo = col4_lo;
    key->col_output[5].lo = col5_lo;
    key->col_output[6].lo = col6_lo;
    key->col_output[7].lo = col7_lo;
    
    key->row_input[0].val = row0_val;
    key->row_input[1].val = row1_val;
    key->row_input[2].val = row2_val;
    key->row_input[3].val = row3_val;
    key->row_input[4].val = row4_val;
    key->row_input[5].val = row5_val;
    key->row_input[6].val = row6_val;
    key->row_input[7].val = row7_val;
    
    key->hw_config = noah_key_hw_config;
        
    for (key->col = 0; key->col < NOAH_KEY_NR_COL; key->col++) {
        for (key->row = 0; key->row < NOAH_KEY_NR_ROW; key->row++) {        
            key->state[key->col][key->row].pressed = 0;
            key->state[key->col][key->row].debounce = 0;
            key->state[key->col][key->row].jif = 0;
        }
    }
    
    key->buf.count = 0;
    key->buf.head = 0;
    key->buf.tail = 0;
 
    spin_lock_init(&key->lock);
    init_waitqueue_head(&key->wait_queue); 
    
    key->poll_timer.function = noah_key_timer_dispatch;
    key->poll_timer.data = (int)key;
    key->poll_timer.expires = jiffies + NOAH_KEY_POLL_INTERVAL;
    add_timer(&key->poll_timer);
}

static void noah_key_ctx_cleanup (struct noah_key_t *key)
{
    dprintk("%s()\n", __FUNCTION__);
    
    del_timer(&key->poll_timer);
}

static struct file_operations noah_key_fops = 
{
    open:       noah_key_open,
    release:    noah_key_release,
    read:       noah_key_read,
    poll:       noah_key_poll,
};

static void __init noah_key_hw_init (struct noah_key_t *key)
{
    int col;
    
    dprintk("%s()\n", __FUNCTION__);
        
    key->hw_config();
    for (col = 0; col < NOAH_KEY_NR_COL; col++)
        key->col_output[col].lo();
}

int __init noah_key_init (void)
{
    int err;
    
    if ((err = register_chrdev(NOAH_KEY_MAJOR, __driver_name__, &noah_key_fops))) {
		printk("Error: unable to register driver %s, err %d\n", __driver_name__, err);
        return err;
    }        

    noah_key_ctx_init(&noah_key);
    noah_key_hw_init(&noah_key);

	printk(KERN_INFO "%s installed\n", __description__);

	return 0;
}

void __exit noah_key_exit (void)
{
    noah_key_ctx_cleanup(&noah_key);
    
    unregister_chrdev(NOAH_KEY_MAJOR, __driver_name__);

    printk(KERN_INFO "%s removed\n", __description__);
}

MODULE_AUTHOR(__author__);
MODULE_DESCRIPTION(__description__);
MODULE_LICENSE(__license__);

module_init(noah_key_init);
module_exit(noah_key_exit);
