/*
 * 	2004-05-06 Sam Cheng (sam@emsoftltd.com)
 * 	s3c44b0 framebuffer driver for GPM425A0 LCD module
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/s3c44b0x_lcd.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/arch/io.h>
#include <asm/arch/s3c44b0x.h>
#include <asm/arch/s3c44b0x_iodirect.h>
#include <asm/arch/memory.h>

#define __driver_name__ "s3c44b0x_lcd"

#define dprintk(a...)		//printk(##a)

//#define S3C44B0X_LCD_DRAW_LOGO
//#define S3C44B0X_LCD_DRAW_TEST_PATTERN

#ifdef S3C44B0X_LCD_DRAW_LOGO
#include "s3c44b0x_lcd_logo.h"
#endif

#define LCD_XSIZE 720
#define LCD_YSIZE 160

#define SCR_XSIZE LCD_XSIZE
#define SCR_YSIZE LCD_YSIZE

#define M5D(n) ((n) & 0x1fffff)

#define HOZVAL			        (LCD_XSIZE / 4 - 1)
#define HOZVAL_COLOR	    	(LCD_XSIZE / 8 - 1)
#define LINEVAL			        (LCD_YSIZE - 1)
#define MVAL			        (13)		
#define CLKVAL_COLOR 	    	(13)		
#define MVAL_USED 0

#define S3C44B0X_FB_ADDRESS         VIDEO_MEM   // calculated and defined in asm/arch/memory.h
#define S3C44B0X_FULL_LEVEL         2500
#define S3C44B0X_DEFAULT_BACKLIGHT  (S3C44B0X_FULL_LEVEL - 100)
#define S3C44B0X_DEFAULT_CONTRAST   3

struct s3c44b0x_lcd_t
{
    unsigned short backlight;
    unsigned short contrast;
    unsigned long power;
    unsigned long *frame_buffer;
};

static struct s3c44b0x_lcd_t s3c44b0x_lcd;

static void s3c44b0x_lcd_init_pwm (void)
{    
    // disable timer 1 and 2
    IO_S3C44B0X_TCON &= ~((0xF << 12) | (0xF << 8));
        
    // contrast PWM, each cycle is 2000Hz        
    IO_S3C44B0X_TCNTB2 = S3C44B0X_FULL_LEVEL;
    IO_S3C44B0X_TCMPB2 = s3c44b0x_lcd.contrast;  
    
    // Black light PWM, each cycle is 2000Hz
    IO_S3C44B0X_TCNTB1 = S3C44B0X_FULL_LEVEL;
    IO_S3C44B0X_TCMPB1 = s3c44b0x_lcd.backlight;
    
    // manual update timer 1 and 2
    IO_S3C44B0X_TCON |= (1 << 13) | (1 << 9);
    
    // turn off manual update
    IO_S3C44B0X_TCON &= ~((1 << 13) | (1 << 9));
    
    // start timer 1 and 2, interval mode, auto reload
    IO_S3C44B0X_TCON |= (0x9 << 12) | (0x9 << 8);
}

#ifdef S3C44B0X_LCD_DRAW_LOGO
static void s3c44b0x_lcd_draw_logo (void) 
{
    unsigned char *addr;

    // disable lcd controller
	IO_S3C44B0X_LCDCON1 &= ~0x1;
    
	addr = (unsigned char *)s3c44b0x_lcd.frame_buffer;
	memcpy(addr, s3c44b0x_lcd_logo, VIDEO_SIZE);
    
    // enable lcd controller
	IO_S3C44B0X_LCDCON1 |= 0x1;
}
#endif

#ifdef S3C44B0X_LCD_DRAW_TEST_PATTERN
static void s3c44b0x_lcd_draw_test_pattern (void) 
{
    unsigned char *addr;
	int i;

    // disable lcd controller
	IO_S3C44B0X_LCDCON1 &= ~0x1;
    
	addr = (unsigned char *)s3c44b0x_lcd.frame_buffer;
	for (i = 0; i <120*40 ; i++) {
		*addr = 0xF0;
		addr++;
		*addr = 0x0F;
		addr++;
		*addr = 0x00;
		addr++;
	}
	for (i = 0; i <120*40 ; i++) {
		*addr = 0x0F;
		addr++;
		*addr = 0x00;
		addr++;
		*addr = 0xF0;
		addr++;
	}
	for (i = 0; i <120*40 ; i++) {
		*addr = 0x00;
		addr++;
		*addr = 0xF0;
		addr++;
		*addr = 0x0F;
		addr++;
	}
	for (i = 0; i <120*20 ; i++) {
		*addr = 0xFF;
		addr++;
		*addr = 0xFF;
		addr++;
		*addr = 0xFF;
		addr++;
	}
	for (i = 0; i <120*20 ; i++) {
		*addr = 0x00;
		addr++;
		*addr = 0x00;
		addr++;
		*addr = 0x00;
		addr++;
	}
	
    // enable lcd controller
	IO_S3C44B0X_LCDCON1 |= 0x1;
}
#endif

static void s3c44b0x_lcd_hw_init (void)
{
    unsigned long lcdcon1, lcdcon2, lcdaddr1, lcdaddr2, lcdaddr3;

    // enable clock into LCD controller
    IO_S3C44B0X_CLKCON |= (1 << 6);

    // port D for video
    IO_S3C44B0X_PCOND = 0xAAAA;

    // pull up disable
    IO_S3C44B0X_PUPD = 0xFF;

	// PE 4 and 5 as PWM timer 1 and 2
    IO_S3C44B0X_PCONE &= ~(0xF << 8);
    IO_S3C44B0X_PCONE |=  (0xA << 8);
    
    // GPC[4:7] => VD[7:4]
    IO_S3C44B0X_PCONC |= (0xFF << 8); 
    // GPC13 output
    IO_S3C44B0X_PCONC &= ~(3 << 26);
    IO_S3C44B0X_PCONC |=  (1 << 26);
    
    // GPC8 output
    IO_S3C44B0X_PCONC &= ~(3 << 16);
    IO_S3C44B0X_PCONC |=  (1 << 16);
    
    // GPC13 hi, DC-DC on
    IO_S3C44B0X_PDATC |= (1 << 13) | (1 << 8);

	lcdcon1 = (0) | (2 << 5) | (MVAL_USED << 7) | (0x3 << 8) | (0x3 << 10) | (CLKVAL_COLOR << 12);    
    // disable, 8B_SNGL_SCAN, WDLY = 8clk, WLH = 8clk,
	lcdcon2 = (LINEVAL) | (HOZVAL_COLOR << 10)| (10 << 21);    
	lcdaddr1 = (0x2 << 27) | (((unsigned long)s3c44b0x_lcd.frame_buffer >> 22)  << 21) | M5D((unsigned long)s3c44b0x_lcd.frame_buffer >> 1);
    // 256-color, LCDBANK, LCDBASEU
	lcdaddr2 = M5D((((unsigned long)s3c44b0x_lcd.frame_buffer + (SCR_XSIZE / 2 * LCD_YSIZE)) >> 1)) | (MVAL << 21) | (1 << 29);
	lcdaddr3 = (LCD_XSIZE / 4) | (((SCR_XSIZE - LCD_XSIZE) / 2) << 9);
    
    outl(lcdcon1, S3C44B0X_LCDCON1);
    outl(lcdcon2, S3C44B0X_LCDCON2);
    outl(lcdaddr1, S3C44B0X_LCDSADDR1);
    outl(lcdaddr2, S3C44B0X_LCDSADDR2);
    outl(lcdaddr3, S3C44B0X_LCDSADDR3);

	//The following value has to be changed for better display.
	IO_S3C44B0X_REDLUT = 0xFDB96420;
	IO_S3C44B0X_GREENLUT = 0xFDB96420;
    IO_S3C44B0X_BLUELUT = 0xFB40;
    
	IO_S3C44B0X_DITHMODE    = 0x0;
	IO_S3C44B0X_DP1_2       = 0xA5A5;
	IO_S3C44B0X_DP4_7       = 0xBA5DA65;
	IO_S3C44B0X_DP3_5       = 0xA5A5f;
	IO_S3C44B0X_DP2_3       = 0xD6B;
	IO_S3C44B0X_DP5_7       = 0xEB7B5ED;
	IO_S3C44B0X_DP3_4       = 0x7DBE;
	IO_S3C44B0X_DP4_5       = 0x7EBDF;
	IO_S3C44B0X_DP6_7       = 0x7FDFBFE;

    // enable LCD controller
	IO_S3C44B0X_LCDCON1 |= 1;
}

static void s3c44b0x_lcd_hw_cleanup (void)
{
    // disable LCD controller
    IO_S3C44B0X_LCDCON1 &= ~(1 << 0);
    
    // GPC13 lo, DC-DC off
    IO_S3C44B0X_PDATC &= ~((1 << 13) | (1 << 8));
    
    // disable clock into LCD controller
    IO_S3C44B0X_CLKCON &= ~(1 << 6);
}

static int s3c44b0x_lcd_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
    unsigned short contrast;
    unsigned short backlight;
    unsigned long power;

    switch (cmd)
    {
        case S3C44B0X_LCD_IOC_GET_FRAMEBUFFER:      
            if (copy_to_user((void *)arg, &s3c44b0x_lcd.frame_buffer, sizeof(unsigned long)))
                return -EFAULT;                
            return 0;
            
        case S3C44B0X_LCD_IOC_GET_CONTRAST:
            if (copy_to_user((void *)arg, &s3c44b0x_lcd.contrast, sizeof(unsigned  short)))
                return -EFAULT;
            return 0;
            
        case S3C44B0X_LCD_IOC_SET_CONTRAST:
            if (copy_from_user(&contrast, (void *)arg, sizeof(unsigned short)))
                return -EFAULT;

            if (contrast <= 10)
            {   
                if (contrast != s3c44b0x_lcd.contrast)
                {
                    s3c44b0x_lcd.contrast = contrast;
                    outw(s3c44b0x_lcd.contrast, S3C44B0X_TCMPB2);
                }
                return 0;
            }   
            return -EINVAL;     

        case S3C44B0X_LCD_IOC_GET_BACKLIGHT:
            if (copy_to_user((void *)arg, &s3c44b0x_lcd.backlight, sizeof(unsigned short)))
                return -EFAULT;
            return 0;            
            
        case S3C44B0X_LCD_IOC_SET_BACKLIGHT:
            if (copy_from_user(&backlight, (void *)arg, sizeof(unsigned short)))
                return -EFAULT;

            if (backlight <= 2400)
            {
                if (backlight != s3c44b0x_lcd.backlight)
                {
                    s3c44b0x_lcd.backlight = backlight;
                    outw(s3c44b0x_lcd.backlight, S3C44B0X_TCMPB1);
                }                    
                return 0;
            }   
            return -EINVAL;
            
        case S3C44B0X_LCD_IOC_GET_POWER_STATE:
            if (copy_to_user((void *)arg, &s3c44b0x_lcd.power, sizeof(unsigned long)))
                return -EFAULT;
            return 0;                
        
        case S3C44B0X_LCD_IOC_SET_POWER_STATE:
            if (copy_from_user(&power, (void *)arg, sizeof(unsigned long)))
                return -EFAULT;
                
            if (power)
            {
                if (!s3c44b0x_lcd.power)
                {
                    s3c44b0x_lcd.power = 1;
                    // power on
                    s3c44b0x_lcd_hw_init();
	                s3c44b0x_lcd_init_pwm();
                }
            }
            else
            {
                if (s3c44b0x_lcd.power)
                {
                    s3c44b0x_lcd.power = 0;
                    // power off
                    s3c44b0x_lcd_hw_cleanup();
                }
            }
            return 0;
                
        default:        
            return -EINVAL;
    }                
}

static ssize_t s3c44b0x_lcd_write (struct file * file, const char *buf, size_t count, loff_t *ppos)
{
    int size;
    
    if ((count + *ppos) > VIDEO_SIZE)
        size = VIDEO_SIZE - *ppos;
    else
        size = count;
    
    if (size == 0)
        return -ENOSPC;
                
	copy_from_user((unsigned char *)s3c44b0x_lcd.frame_buffer + *ppos, buf, size);
    *ppos += size;
    
    return size;
}

static struct file_operations s3c44b0x_lcd_fops =
{
    write:      s3c44b0x_lcd_write,
    ioctl:      s3c44b0x_lcd_ioctl,
};

static __init int s3c44b0x_lcd_init(void)
{
    s3c44b0x_lcd.frame_buffer = (unsigned long *)S3C44B0X_FB_ADDRESS;
    s3c44b0x_lcd.backlight = S3C44B0X_DEFAULT_BACKLIGHT;
    s3c44b0x_lcd.contrast = S3C44B0X_DEFAULT_CONTRAST;
    s3c44b0x_lcd.power = 1;

	s3c44b0x_lcd_hw_init();
	s3c44b0x_lcd_init_pwm();

	if (register_chrdev(101, __driver_name__, &s3c44b0x_lcd_fops)) 
		panic("Unable to register %s\n", __driver_name__);
	
    memset(s3c44b0x_lcd.frame_buffer, 0xff, VIDEO_SIZE);
    
#ifdef S3C44B0X_LCD_DRAW_LOGO    
    s3c44b0x_lcd_draw_logo();
#endif

#ifdef S3C44B0X_LCD_DRAW_TEST_PATTERN
    s3c44b0x_lcd_draw_test_pattern();    
#endif
    
	printk(KERN_INFO "Samsung s3c44b0x 4096 color STN driver frame buffer: 0x%08X\n", (int)s3c44b0x_lcd.frame_buffer);

	return 0;
}

static __exit void s3c44b0x_lcd_exit(void)
{
    unregister_chrdev(101, __driver_name__);
}

module_init(s3c44b0x_lcd_init);
module_exit(s3c44b0x_lcd_exit);
